from abc import ABC
from typing import Optional


class AbstractClient(ABC):
    api_key: str
    base_url: str
    model: Optional[str]
    messages = []

    def as_chat(self, model: str, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.messages = [{"role": "system", "content": sys_prompt}]

    def as_embeddings(self, model: str):
        self.model = model

    def chat(self, *user_prompt, **kwargs):
        ...

    def encode(self, *_input: str) -> list[list[float]]:
        ...
    def connect(self):
        ...

    def disconnect(self):
        ...



import json
from typing import TypedDict

import requests
from davidkhala.utils.http_request.stream import Request as StreamRequest, as_sse
from requests import Response, Session

from davidkhala.ai.agent.dify.api import API


class Feedbacks(API):
    def paginate_feedbacks(self, page=1, size=20):
        """
        when 'rating'='like', content=None
        when 'rating'='dislike', content can be filled by end user
        NOTE: for security reason, api cannot access conversation context associated with the feedback. End user should copy the conversation to comment by themselves.
        # waiting for https://github.com/langgenius/dify/issues/28067
        """
        response = requests.get(f"{self.base_url}/app/feedbacks", params={"page": page, "limit": size}, **self.options)
        if not response.ok:
            response.raise_for_status()
        else:
            return json.loads(response.text)

    def list_feedbacks(self):
        return self.paginate_feedbacks()['data']


class Conversation(API):
    """
    Note: The Service API does not share conversations created by the WebApp. Conversations created through the API are isolated from those created in the WebApp interface.
    It means you cannot get user conversation content from API, API call has only access to conversation created by API
    """

    def __init__(self, api_key: str, user: str):
        super().__init__(api_key)  # base_url need to be configured afterward if not default
        self.user = user  # user_id, from_end_user_id

    def paginate_messages(self, conversation_id):
        return self.request(f"{self.base_url}/messages", "GET", params={
            'conversation_id': conversation_id,
            'user': self.user,
        })

    def _chat_request_from(self, template: str, stream, **kwargs):
        """
        :param template:
        :param stream: Note: "Agent Chat App does not support blocking mode"
        :param kwargs:
        :return:
        """
        return {
            'url': f"{self.base_url}/chat-messages",
            'method': "POST",
            'json': {
                'query': template,
                'inputs': kwargs.pop('values', {}),  # to substitute query/template
                'response_mode': 'streaming' if stream else 'blocking',
                'conversation_id': kwargs.pop('conversation_id', None),
                'user': self.user,
                'files': kwargs.pop('files', [])
            },
            **kwargs
        }

    def async_chat(self, template: str, **kwargs) -> tuple[Response, Session]:
        s = StreamRequest(self)
        s.session = Session()
        return s.request(**self._chat_request_from(template, True, **kwargs)), s.session

    class ChatResult(TypedDict, total=False):
        thought: list[str]
        metadata: dict

    @staticmethod
    def reduce_chat_stream(response: Response) -> ChatResult:
        r: Conversation.ChatResult = {
            'thought': [],
        }
        for data in as_sse(response):
            match data['event']:
                case 'agent_thought':
                    r['thought'].append(data['thought'])
                case 'message_end':
                    r['metadata'] = data['metadata']
        return r

    def agent_chat(self, template: str, **kwargs) -> ChatResult:
        r, session = self.async_chat(template, **kwargs)
        reduced = Conversation.reduce_chat_stream(r)
        session.close()
        return reduced

    def bot_chat(self, template: str, **kwargs):
        r = self.request(**self._chat_request_from(template, False, **kwargs))
        assert r.pop('event') == 'message'
        assert r.pop('mode') == 'chat'
        return r



from typing import Iterable, Callable, Any, Optional

from davidkhala.utils.http_request import Request


class API(Request):
    def __init__(self, api_key: str, base_url="https://api.dify.ai/v1"):
        super().__init__({'bearer': api_key})
        self.base_url = base_url
        self.api_key = api_key


class Iterator(Iterable):
    def __iter__(self):
        return self

    def __init__(self, get_fn: Callable[[int, int], Any], r: Optional[dict]):
        self.response = r
        self.fn = get_fn

    def __next__(self):
        if self.response and not self.response['has_more']:
            raise StopIteration
        page = 1 if not self.response else self.response['page'] + 1
        limit = None if not self.response else self.response['limit']
        self.response = self.fn(page, limit)
        return self.response['data']



from typing import Optional, Literal

from openai import OpenAI

from davidkhala.ai.openai import Client


class NativeClient(Client):
    def __init__(self, api_key, base_url=None):
        self.client = OpenAI(
            api_key=api_key,
            base_url=base_url
        )

    def chat(self, *user_prompt, web_search:Optional[Literal["low", "medium", "high"]]=None, **kwargs):
        opts = {
            **kwargs
        }
        if web_search:
            from openai.types.chat.completion_create_params import WebSearchOptions
            opts['web_search_options'] = WebSearchOptions(search_context_size=web_search)
        return super().chat(*user_prompt, **opts)



from openrouter.errors import UnauthorizedResponseError

from davidkhala.ai.model import AbstractClient
from openrouter import OpenRouter


class Client(AbstractClient):
    def __init__(self, api_key: str):
        self.api_key = api_key
        self.client = OpenRouter(api_key)

    def chat(self, *user_prompt, **kwargs):
        r = self.client.chat.send(
            model=self.model,
            messages=[
                *self.messages,
                *[{'role': 'user', 'content': _} for _ in user_prompt]
            ]
        )
        return [_.message.content for _ in r.choices]
    def connect(self):
        try:
            self.client.models.list()
            return True
        except UnauthorizedResponseError:
            return False


class Admin:
    def __init__(self, provisioning_key: str):
        self.provisioning_key = provisioning_key
        self.client = OpenRouter(provisioning_key)
    @property
    def keys(self):
        return self.client.api_keys.list().data


from davidkhala.utils.http_request import Request


class API(Request):
    def __init__(self, base_url='http://localhost'):
        super().__init__()
        self.base_url = f"{base_url}/console/api"
        self.__enter__()




from davidkhala.ai.agent.dify.ops.db import DB
from davidkhala.ai.agent.dify.ops.db.orm import Graph


class Dataset(DB):

    def dataset_queries(self, dataset_id, limit=20) -> list[str]:
        template = "select content from dataset_queries where source = 'app' and created_by_role = 'end_user' and dataset_id = :dataset_id limit :limit"
        return self.query(template, {'dataset_id': dataset_id, 'limit': limit}).scalars().all()

    @property
    def datasets(self):
        template = "select id, name, description, indexing_technique, index_struct, embedding_model, embedding_model_provider, collection_binding_id, retrieval_model, icon_info, runtime_mode, pipeline_id, chunk_structure from datasets"
        return self.get_dict(template)

    def is_pipeline(self, id: str):
        template = "select runtime_mode = 'rag_pipeline' from datasets where id = :id"
        return self.query(template, {'id': id}).scalar()

    @property
    def data_source_credentials(self):
        template = "select id, name, plugin_id, auth_type from datasource_providers"
        return self.get_dict(template)

    def credential_id_by(self, name, provider) -> list[str]:
        template = "select id from datasource_providers where name = :name and provider = :provider"
        return self.query(template, {'name': name, 'provider': provider}).scalars().all()

    def documents(self, dataset_id: str):
        template = "select id, name,created_from, created_at from documents where dataset_id = :dataset_id"
        return self.query(template, {'dataset_id': dataset_id})


class Document(DB):
    def hit_documents(self, top_k: int = 3):
        template = "SELECT dataset_id, document_id, content FROM document_segments ORDER BY hit_count DESC LIMIT :top_k"
        return self.get_dict(template, {'top_k': top_k})

    def id_by(self, name: str, dataset_id: str = None) -> list[str]:
        """multiple ids can be found"""
        template = "select id from documents where name = :name"
        if dataset_id:
            template = "select id from documents where name = :name and dataset_id = :dataset_id"
        return [str(uuid) for uuid in self.query(template, {'name': name, 'dataset_id': dataset_id}).scalars().all()]


class Pipeline(DB):
    @property
    def pipelines(self):
        """unique syntax for pgsql"""
        template = "SELECT DISTINCT ON (app_id) app_id, graph, rag_pipeline_variables FROM workflows where type = 'rag-pipeline' ORDER BY app_id, created_at DESC"
        return Graph.convert(*self.get_dict(template))

    def pipeline(self, app_id):
        template = "select id, graph, rag_pipeline_variables from workflows where type = 'rag-pipeline' and app_id = :app_id"
        dict_result = self.get_dict(template, {'app_id': app_id})
        assert len(dict_result) < 2
        return Graph.convert(*dict_result)



from __future__ import annotations

import os
from pathlib import Path
from typing import Iterable, TypedDict, Optional
from urllib.parse import urlparse

import requests

from davidkhala.ai.agent.dify.api import API, Iterator
from davidkhala.ai.agent.dify.model import Document as DocumentBase


class DatasetDict(TypedDict):
    id: str
    name: str
    description: str
    provider: str
    permission: str
    data_source_type: str
    indexing_technique: str
    doc_form: str
    runtime_mode: str
    is_published: bool
    enable_api: bool
    # stats
    app_count: int
    document_count: int
    word_count: int
    total_documents: int
    total_available_documents: int
    # embedding
    embedding_available: bool
    embedding_model: str
    embedding_model_provider: str
    retrieval_model_dict: dict
    external_retrieval_model: dict
    external_knowledge_info: dict


class Document(DocumentBase):
    data_source_info: dict[str, str]
    data_source_detail_dict: dict[str, dict]
    dataset_process_rule_id: str
    created_from: str
    created_by: str
    created_at: int
    tokens: int
    archived: bool
    display_status: str
    word_count: int
    hit_count: int
    doc_form: str
    doc_metadata: dict
    disabled_at: int
    disabled_by: str


class Dataset(API):
    def __init__(self, api_key: str, base_url="https://api.dify.ai/v1"):
        super().__init__(api_key, f"{base_url}/datasets")

    def paginate_datasets(self, page=1, size=20):
        r = self.request(self.base_url, "GET", params={
            'page': page,
            'limit': size,
        })
        return r

    def list_datasets(self) -> Iterable[list[DatasetDict]]:
        return Iterator(self.paginate_datasets, None)

    @property
    def ids(self):
        for sub_list in self.list_datasets():
            for dataset in sub_list:
                yield dataset['id']

    class Instance(API):
        def __init__(self, d: Dataset, dataset_id: str):
            super().__init__(d.api_key, f"{d.base_url}/{dataset_id}")

        def get(self):
            return self.request(self.base_url, "GET")

        def upload(self, filename, *, path=None, url=None, document_id=None):
            """
            don't work for .html
            work for .md
            """
            files = {}
            if path:
                with open(path, 'rb') as f:
                    content = f.read()
                if not filename:
                    filename = os.path.basename(path)
            elif url:
                r = requests.get(url)
                r.raise_for_status()
                if not filename:
                    parsed_url = urlparse(url)
                    filename = Path(parsed_url.path).name
                content = r.content
            files['file'] = (filename, content)
            if document_id:
                # don't work for html
                r = requests.post(f"{self.base_url}/documents/{document_id}/update-by-file", files=files,
                                  **self.options)
            else:
                r = requests.post(f"{self.base_url}/document/create-by-file", files=files, **self.options)
            r = self.on_response(r)
            return r['document']

        def paginate_documents(self, page=1, size=20):
            return self.request(f"{self.base_url}/documents", "GET", params={
                'page': page,
                'limit': size
            })

        def list_documents(self) -> Iterable[Document]:
            for document_batch in Iterator(self.paginate_documents, None):
                for document in document_batch:
                    yield Document(**document)

        def has_document(self, name) -> bool:
            return any(name == item['name'] for row in self.list_documents() for item in row)


class ChunkDict(TypedDict):
    id: str
    position: int
    document_id: str
    content: str
    sign_content: str  # trimmed version of content
    answer: Optional[str]  # only used in QA chunk
    word_count: int
    tokens: int
    keywords: Optional[list[str]]
    index_node_id: str  # chunk 在向量索引中的节点 ID
    index_node_hash: str  # hash of sign_content
    hit_count: int
    enabled: bool
    status: str  # 'completed'
    created_at: int  # timestamp
    updated_at: int  # timestamp
    completed_at: int  # timestamp
    created_by: str  # user id
    child_chunks: list
    error: Optional
    stopped_at: Optional[int]  # timestamp
    disabled_at: Optional[int]  # timestamp


class Document(API):
    def __init__(self, d: Dataset.Instance, document_id: str):
        super().__init__(d.api_key, f"{d.base_url}/documents/{document_id}")

    def exist(self):
        try:
            self.get()
            return True
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return False
            else:
                raise e

    def get(self):
        return self.request(self.base_url, "GET")

    def paginate_chunks(self, page=1, size=20):
        return self.request(f"{self.base_url}/segments", "GET", params={
            'page': page,
            'limit': size
        })

    def list_chunks(self) -> Iterable[ChunkDict]:
        for chunk_batch in Iterator(self.paginate_chunks, None):
            for chunk in chunk_batch:
                yield chunk

    def delete(self):
        if self.exist():
            self.request(self.base_url, "DELETE")
class Chunk(API):
    def __init__(self, d: Document, segment_id: str):
        super().__init__(d.api_key, f"{d.base_url}/segments/{segment_id}")
    def get(self):
        r=  self.request(self.base_url, "GET")
        assert r['doc_form'] # optional value text_model
        return r['data']


from davidkhala.ai.agent.dify.ops.console import API


class ConsoleUser(API):
    def login(self, email, password,
              *,
              remember_me=True,
              language="en-US"
              ):
        url = f"{self.base_url}/login"

        r = self.request(url, "POST", json={
            'email': email,
            'password': password,
            'remember_me': remember_me,
            'language': language,
        })
        assert r == {"result": "success"}
        self.options['headers']['x-csrf-token'] = self.session.cookies.get("csrf_token")
        return self.session.cookies

    @property
    def me(self) -> dict:
        url = f"{self.base_url}/account/profile"
        return self.request(url, "GET")

    @property
    def workspace(self) -> dict:
        url = f"{self.base_url}/features"
        return self.request(url, "GET")



from langchain_openai import ChatOpenAI
from langgraph.prebuilt import create_react_agent


class Agent:

    def __init__(self, model, instruction, *tools):
        self.agent = create_react_agent(
            model=model,
            tools=tools,
            prompt=instruction
        )

    def invoke(self, content):
        return self.agent.invoke({"messages": [{"role": "user", "content": content}]})['messages'][-1]


class OpenRouterModel:
    def __init__(self, api_key, leaderboard: dict = None):
        self.api_key = api_key

        if leaderboard:
            self.headers = {
                "HTTP-Referer": leaderboard['url'],
                "X-Title": leaderboard['name'],
            }

    def init_chat_model(self, model):
        """https://openrouter.ai/docs/community/lang-chain"""
        return ChatOpenAI(
            base_url='https://openrouter.ai/api/v1',
            model=model,
            api_key=self.api_key,
            default_headers=getattr(self, "headers", None)
        )



from davidkhala.utils.syntax.compat import deprecated
from openai import AzureOpenAI, OpenAI

from davidkhala.ai.openai import Client


class AzureHosted(Client):
    def chat(self, *user_prompt, **kwargs):
        if 'web_search_options' in kwargs:
            raise ValueError('Web search options not supported in any models of Azure AI Foundry')
        return super().chat(*user_prompt, **kwargs)

class ModelDeploymentClient(AzureHosted):
    def __init__(self, key, deployment):
        self.client = AzureOpenAI(
            api_version="2024-12-01-preview",  # mandatory
            azure_endpoint=f"https://{deployment}.cognitiveservices.azure.com/",
            api_key=key,
        )


@deprecated("Azure Open AI is deprecated. Please migrate to Microsoft Foundry")
class OpenAIClient(AzureHosted):

    def __init__(self, api_key, project):
        self.client = OpenAI(
            base_url=f"https://{project}.openai.azure.com/openai/v1/",
            api_key=api_key,
        )

    def as_chat(self, model="gpt-oss-120b", sys_prompt: str = None):
        super().as_chat(model, sys_prompt)



from typing import Any

from davidkhala.data.base.pg import Postgres


class DB(Postgres):

    def __init__(self, connection_string: str):
        super().__init__(connection_string)
        self.connect()

    def get_dict(self,
                 template: str,
                 values: dict[str, Any] | None = None,
                 request_options: dict[str, Any] | None = None
                 ) -> list[dict]:
        return Postgres.rows_to_dicts(self.query(template, values, request_options))



from pydantic import BaseModel, Field

from davidkhala.ai.agent.dify.const import IndexingStatus


class Document(BaseModel):
    id: str
    position: int
    data_source_type: str
    data_source_info: dict[str, str]
    name: str
    indexing_status: IndexingStatus
    error: str | None
    enabled: bool


class Dataset(BaseModel):
    id: str
    name: str
    description: str


class JsonData(BaseModel):
    data: list


class NodeOutput(BaseModel):
    """Schema for Output of a Dify node"""
    text: str
    files: list
    json_: list[JsonData] = Field(alias="json") # avoid conflict with .json()



from huggingface_hub import InferenceApi


class API:
    def __init__(self, token):
        self.inference = None
        self.token = token

    def as_model(self, repo_id):
        self.inference = InferenceApi(repo_id=repo_id, token=self.token)

    def call(self, **kwargs):
        return self.inference(**kwargs)



from agentbay import AgentBay, Session, Config, AgentBayLogger, BrowserOption
from davidkhala.utils.syntax.interface import ContextAware

AgentBayLogger.setup(level='WARNING')  # Default to INFO


class Client(ContextAware):
    def __init__(self, api_key, *, timeout_ms=10000):
        self.agent = AgentBay(
            api_key=api_key,
            cfg=Config(endpoint="wuyingai.ap-southeast-1.aliyuncs.com", timeout_ms=timeout_ms)
        )
        self.session: Session | None = None

    def open(self):
        r = self.agent.create()
        if not r.success:
            return False
        self.session = r.session
        return True

    def close(self):
        self.agent.delete(self.session)
        del self.session


class Browser(ContextAware):
    def __init__(self, session: Session):
        self.session = session
        self.option = BrowserOption()
        self.endpoint_url: str | None = None

    def open(self) -> bool:
        success = self.session.browser.initialize(self.option)
        self.endpoint_url = self.session.browser.get_endpoint_url()
        return success

    def close(self):
        self.session.browser.destroy()


from time import sleep

from davidkhala.utils.http_request.stream import as_sse, Request as StreamRequest
from pydantic import BaseModel

from davidkhala.ai.agent.dify.interface import IndexingError
from davidkhala.ai.agent.dify.model import Document, Dataset
from davidkhala.ai.agent.dify.const import IndexingStatus
from davidkhala.ai.agent.dify.ops.console import API
from davidkhala.ai.agent.dify.ops.console.session import ConsoleUser
from davidkhala.ai.agent.dify.ops.db.orm import Node


class ConsoleKnowledge(API):
    def __init__(self, context: ConsoleUser):
        super().__init__()
        self.base_url = context.base_url
        self.session.cookies = context.session.cookies
        self.options = context.options


class Datasource(ConsoleKnowledge):
    """step 1: Choose a Data Source"""

    class FirecrawlOutput(BaseModel):
        source_url: str
        description: str
        title: str
        credential_id: str
        content: str

    def run_firecrawl(self, pipeline: str, node: Node,
                      *,
                      inputs: dict,
                      credential_id: str
                      ):

        url = f"{self.base_url}/rag/pipelines/{pipeline}/workflows/published/datasource/nodes/{node.id}/run"

        stream_request = StreamRequest(self)
        response = stream_request.request(url, 'POST', json={
            'inputs': inputs,
            'datasource_type': node.datasource_type,
            'credential_id': credential_id,
            "response_mode": "streaming"
        })

        for data in as_sse(response):
            event = data['event']
            if event == 'datasource_completed':
                return data['data']
            else:
                assert event == 'datasource_processing'
                print(data)
        return None

    def upload(self):
        "http://localhost/console/api/files/upload?source=datasets"
        # TODO
        "form data"
        {
            "file": "body"
        }
        r = {
            "id": "3898db5b-eb72-4f11-b507-628ad5d28887",
            "name": "Professional Diploma Meister Power Electrical Engineering - Technological and Higher Education Institute of Hong Kong.html",
            "size": 254362,
            "extension": "html",
            "mime_type": "text\/html",
            "created_by": "dbd0b38b-5ef1-4123-8c3f-0c82eb1feacd",
            "created_at": 1764943811,
            "source_url": "\/files\/3898db5b-eb72-4f11-b507-628ad5d28887\/file-preview?timestamp=1764943811&nonce=43b0ff5a13372415be79de4cc7ef398c&sign=7OJ2wiVYc4tygl7yvM1sPn7s0WXDlhHxgX76bsGTD94%3D"
        }


class Operation(ConsoleKnowledge):
    def website_sync(self, dataset: str, document: str, *, wait_until=True):
        """
        cannot be used towards a pipeline dataset. Otherwise, you will see error "no website import info found"
        """
        doc_url = f"{self.base_url}/datasets/{dataset}/documents/{document}"

        r = self.request(f"{doc_url}/website-sync", "GET")
        assert r == {"result": "success"}
        if wait_until:
            return self.wait_until(dataset, document)
        return None

    def retry(self, dataset: str, *documents: str, wait_until=True):
        """
        It cannot trigger rerun on success documents
        """
        url = f"{self.base_url}/datasets/{dataset}/retry"
        self.request(url, "POST", json={
            'document_ids': documents,
        })
        # response status code will be 204
        if wait_until:
            return [self.wait_until(dataset, document) for document in documents]
        return None

    def rerun(self, dataset: str, *documents: str):
        for document in documents:
            try:
                self.website_sync(dataset, document)
                assert False, "expect IndexingError"
            except IndexingError:
                pass
        return self.retry(dataset, *documents)

    def wait_until(self, dataset: str, document: str, *,
                   expect_status=None,
                   from_status=None,
                   interval=1
                   ):
        if not expect_status:
            expect_status = [IndexingStatus.FAILED, IndexingStatus.COMPLETED]
        url = f"{self.base_url}/datasets/{dataset}/documents/{document}/indexing-status"
        if from_status is None:
            from_status = [IndexingStatus.WAITING, IndexingStatus.PARSING]
        r = self.request(url, "GET")
        status = r['indexing_status']
        assert status in from_status, f"current status: {status}, expect: {from_status}"
        while status not in expect_status:
            sleep(interval)
            r = self.request(url, "GET")
            status = r['indexing_status']
        if status == IndexingStatus.FAILED: raise IndexingError(r['error'])
        return r


class DatasetResult(Dataset):
    chunk_structure: str

class RunResult(BaseModel):
    batch: str
    dataset: DatasetResult
    documents: list[Document]

class Load(ConsoleKnowledge):
    """
    Processing Documents
    """

    def async_run(self, pipeline: str, node: Node, inputs: dict, datasource_info_list: list[dict])->RunResult:
        """Ingest new document"""
        url = f"{self.base_url}/rag/pipelines/{pipeline}/workflows/published/run"
        r = self.request(url, "POST", json={
            'inputs': inputs,
            'start_node_id': node.id,
            'is_preview': False,
            'response_mode': "blocking",
            "datasource_info_list": datasource_info_list,
            'datasource_type': node.datasource_type
        })
        return RunResult(**r)





from pydantic import BaseModel

from davidkhala.ai.agent.dify.plugins import DataSourceTypeAware


class FileModel(BaseModel):
    name: str
    size: int
    type: str
    extension: str
    mime_type: str
    transfer_method: str
    url: str
    related_id: str


class DataSourceOutput(DataSourceTypeAware):
    datasource_type:str = "local_file"
    file: FileModel



import json
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel
from sqlalchemy import Column, String, Text, JSON, TIMESTAMP, func
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class DifyBase(Base):
    __abstract__ = True  # keyword for SQLAlchemy
    id = Column(UUID(as_uuid=True), primary_key=True, server_default=func.uuid_generate_v4())


class AppModelConfig(DifyBase):
    __tablename__ = "app_model_configs"
    __table_args__ = {"schema": "public"}

    app_id = Column(UUID(as_uuid=True), nullable=False)

    provider = Column(String(255))
    model_id = Column(String(255))
    configs = Column(JSON)

    created_at = Column(TIMESTAMP, nullable=False, server_default=func.current_timestamp())
    updated_at = Column(TIMESTAMP, nullable=False, server_default=func.current_timestamp())

    opening_statement = Column(Text)
    suggested_questions = Column(Text)
    suggested_questions_after_answer = Column(Text)
    more_like_this = Column(Text)
    model = Column(Text)
    user_input_form = Column(Text)
    pre_prompt = Column(Text)
    agent_mode = Column(Text)
    speech_to_text = Column(Text)
    sensitive_word_avoidance = Column(Text)
    retriever_resource = Column(Text)

    dataset_query_variable = Column(String(255))
    prompt_type = Column(String(255), nullable=False, server_default="simple")

    chat_prompt_config = Column(Text)
    completion_prompt_config = Column(Text)
    dataset_configs = Column(Text)
    external_data_tools = Column(Text)
    file_upload = Column(Text)
    text_to_speech = Column(Text)

    created_by = Column(UUID(as_uuid=True))
    updated_by = Column(UUID(as_uuid=True))

    def __repr__(self):
        return f"<AppModelConfig(id={self.id}, app_id={self.app_id}, provider={self.provider}, model_id={self.model_id})>"


class Position(BaseModel):
    x: float
    y: float


class NodeData(BaseModel):
    class Type(str, Enum):
        SOURCE = 'datasource'
        CHUNKER = 'knowledge-index'
        TOOL = 'tool'

    type: Type | str  # not limit to built-in types
    title: str | None = None
    selected: bool

    # datasource
    datasource_parameters: dict[str, Any] | None = None
    datasource_configurations: dict[str, Any] | None = None
    plugin_id: str | None = None
    provider_type: str | None = None
    provider_name: str | None = None
    datasource_name: str | None = None
    datasource_label: str | None = None
    plugin_unique_identifier: str | None = None

    # tool
    tool_parameters: dict[str, Any] | None = None
    tool_configurations: dict[str, Any] | None = None
    tool_node_version: str | None = None
    provider_id: str | None = None
    provider_icon: str | None = None
    tool_name: str | None = None
    tool_label: str | None = None
    tool_description: str | None = None
    is_team_authorization: bool | None = None
    paramSchemas: list[Any] | None = None
    params: dict[str, Any] | None = None

    # knowledge index
    index_chunk_variable_selector: list[str] | None = None
    keyword_number: int | None = None
    retrieval_model: dict[str, Any] | None = None
    chunk_structure: str | None = None
    indexing_technique: str | None = None
    embedding_model: str | None = None
    embedding_model_provider: str | None = None


class Node(BaseModel):
    @property
    def datasource_type(self): return self.data.provider_type
    id: str
    type: Literal['custom']
    data: NodeData
    position: Position
    targetPosition: str | None = None
    sourcePosition: str | None = None
    positionAbsolute: Position | None = None
    width: float | None = None
    height: float | None = None
    selected: bool


class Edge(BaseModel):
    id: str
    type: str
    source: str
    target: str
    sourceHandle: str | None = None
    targetHandle: str | None = None
    data: dict[str, Any] | None = None
    zIndex: int | None = None


class Viewport(BaseModel):
    x: float
    y: float
    zoom: float


class Graph(BaseModel):
    nodes: list[Node]
    edges: list[Edge]
    viewport: Viewport

    @property
    def datasources(self):
        return [node for node in self.nodes if node.data.type == NodeData.Type.SOURCE]

    @staticmethod
    def convert(*records: list[dict]):
        return [{**record, "graph": Graph(**json.loads(record["graph"]))} for record in records]
